# Copyright (C) 2020, The Beanstalks Project ehf. and Bjarni R. Einarsson.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Commercial licenses are for sale. See the files README.md and COPYING.txt
# for more details.


def xml_encode(s):
  return (str(s)
    .replace('&', '&amp;')
    .replace('"', '&quot;')
    .replace('<', '&lt;')
    .replace('>', '&gt;'))


def to_xml(name, data, indent='', attrs=''):
  if isinstance(data, dict):
    if 'value' in data:
      attrs = data
      data = attrs['value']
      del attrs['value']
    elif '_xml_attrs' in data:
      attrs = data['_xml_attrs']
      del data['_xml_attrs']
    if '_xml_name' in attrs:
      name = attrs['_xml_name']
      del attrs['_xml_name']

  if isinstance(data, dict):
    data = '\n'.join(sorted([to_xml(k, data[k], indent+' ') for k in data]))
    if data:
      data = '\n%s\n%s' % (data, indent)
  elif isinstance(data, list):
    ename = name[:-1] if (name[-1:] == 's') else 'item'
    data = '\n'.join([to_xml(ename, d, indent+' ') for d in data])
    if data:
      data = '\n%s\n%s' % (data, indent)
  else:
    data = xml_encode(data)

  if isinstance(attrs, dict):
    attrs = ' '.join(sorted([
      '%s="%s"' % (k, xml_encode(attrs[k])) for k in attrs]))

  if indent:
    fmt = indent + '<%s%s%s>%s</%s>'
  else:
    fmt = '<?xml version="1.0"?>\n<%s%s%s>%s</%s>\n'

  return fmt % (name, ' ' if attrs else '', attrs, data, name)


if False and (__name__ == '__main__'):
  print(to_xml('test', {
    '_xml_attrs': {'fancy': True, 'version': 1},
    'hello': 'world',
    '000': 'Zeroth Element',
    'elements': [1, 2, 3, 4]}))



# Copyright (C) 2020, The Beanstalks Project ehf. and Bjarni R. Einarsson.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Commercial licenses are for sale. See the files README.md and COPYING.txt
# for more details.
#
import sys

from . import Kite, uPageKite, uPageKiteDefaults


class MyDefaults(uPageKiteDefaults):
  # Disable connecting to new front-ends, static setup
  #FE_NAME = None

  # Enable all the logging
  trace = uPageKiteDefaults.log
  debug = uPageKiteDefaults.log
  info  = uPageKiteDefaults.log
  error = uPageKiteDefaults.log


async def handle_http_request(kite, conn, frame):
  await conn.reply(frame, (
      'HTTP/1.0 200 OK\n'
      'Content-Type: text/html\n'
      '\n'
      '<h1>Hello world!</h1><h2>This is %s, you are %s</h2>\n'
    ) % (kite.name, frame.remote_ip))


if __name__ == "__main__":
  try:
    uPageKite([
        Kite(sys.argv[1], sys.argv[2], handler=handle_http_request)
      ], uPK=MyDefaults).run()
  except IndexError:
    print('Usage: %s kitename kitesecret' % sys.argv[0])



# Copyright (C) 2020, The Beanstalks Project ehf. and Bjarni R. Einarsson.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Commercial licenses are for sale. See the files README.md and COPYING.txt
# for more details.

import struct
from hashlib import sha1

from .proto import asyncio, fuzzy_sleep_ms

try:
  from ubinascii import b2a_base64
  def sha1b64(data):
    return str(
      b2a_base64(sha1(bytes(data, 'latin-1')).digest()),
      'latin-1').strip()
except (NameError, ImportError):
  from base64 import b64encode
  def sha1b64(data):
    return str(
      b64encode(sha1(bytes(data, 'latin-1')).digest()),
      'latin-1').strip()


_WEBSOCKETS = {}


# Decorator for creating a websocket and registering the handler for
# incoming messages.
def websocket(ws_id=None, strict_origin=True):
  def decorate(message_handler):
    async def url_handler(req_env):
      hdrs = req_env['http_headers']

      if ((hdrs.get('Upgrade', '').lower() != 'websocket')
          or (not hdrs.get('Sec-WebSocket-Key'))
          or (hdrs.get('Sec-WebSocket-Protocol'))  # Unsupported!
          or (hdrs.get('Sec-WebSocket-Version', '') not in ('13', ))):
        rv = None
        if 'Upgrade' not in hdrs:
          rv = await message_handler(None, None, None, None, websocket=False)
        return rv or {'code': 400, 'msg': 'Invalid Request'}

      if strict_origin and (not hdrs.get('Host')
          or (('://'+hdrs.get('Host', '')) not in hdrs.get('Origin', ''))):
        return {'code': 403, 'msg': 'Forbidden'}

      uPK = req_env['httpd'].uPK
      live_conns = sum([len(ws.streams) for ws in _WEBSOCKETS.values()])
      if live_conns >= uPK.WEBSOCKET_MAX_CONNS:
        return {'code': 503, 'msg': 'Too Many Clients'}

      conn = req_env['conn']
      frame = req_env['frame']

      ws = Websocket.get(ws_id or req_env.frame.uid, message_handler, uPK)
      await ws.subscribe(conn, frame, req_env)

      key = req_env['http_headers']['Sec-WebSocket-Key']
      signature = sha1b64(key + '258EAFA5-E914-47DA-95CA-C5AB0DC85B11')
      return {
        'code': 101,
        'msg': 'Switching Protocols',
        'mimetype': None,
        'hdrs': {
          'Upgrade': 'websocket',
          'Sec-WebSocket-Accept': signature,
          'Connection': 'Upgrade'},
        'eof': False}
    return url_handler
  return decorate


async def ws_broadcast(ws_ids, message, only=None):
  for _id in (ws_ids if isinstance(ws_ids, list) else [ws_ids]):
    try:
      await Websocket.get(_id).broadcast(message, only=only)
    except KeyError:
      pass


class OPCODES(object):
  CONT = 0x0
  TEXT = 0x1
  BINARY = 0x2
  CLOSE = 0x8
  PING = 0x9
  PONG = 0xa


class Websocket(object):
  def __init__(self, ws_id, message_handler, uPK):
    self.uPK = uPK
    self.ws_id = ws_id
    self.message_handler = message_handler
    self.streams = {}
    self.make_mask = uPK.WEBSOCKET_MASK

  @classmethod
  def get(cls, ws_id, message_handler=None, uPK=None):
    global _WEBSOCKETS
    if ws_id not in _WEBSOCKETS:
      if message_handler and uPK:
        _WEBSOCKETS[ws_id] = cls(ws_id, message_handler, uPK)
    return _WEBSOCKETS[ws_id]

  async def subscribe(self, conn, frame, env):
    if self.uPK.info:
      self.uPK.info('[ws/%s] Subscribe %s %s'
        % (self.ws_id, frame.uid, frame.remote_ip))

    wss = WebsocketStream(conn, frame, env, self.uPK, self.make_mask())
    self.streams[frame.uid] = wss
    conn.async_await_data(self.uPK, frame.sid, self.receive_data)

    async def welcome():
      await fuzzy_sleep_ms(25)
      await self.message_handler(None, None, wss, self, first=True)
    asyncio.get_event_loop().create_task(welcome())

  def unsubscribe(self, uid):
    # FIXME: Stop awaiting data
    if uid in self.streams:
      wss = self.streams[uid]
      if self.uPK.info:
        ip = wss.frame.remote_ip
        self.uPK.info('[ws/%s] Unsubscribe %s %s' % (self.ws_id, uid, ip))
      wss.conn.close(wss.frame.sid)
      del self.streams[uid]

  async def receive_data(self, frame):
    if frame.headers.get('NOOP'):
      return

    wss = self.streams.get(frame.uid)
    if not wss:
      return  # FIXME: Send back an EOF

    if frame.eof:
      self.unsubscribe(frame.uid)
      await self.message_handler(None, None, wss, self, eof=True)
    else:
      for opcode, message in wss.process(frame.payload):
        if opcode == OPCODES.TEXT:
          message = str(message, 'utf-8')
        if self.uPK.trace:
          self.uPK.trace('[ws/%s] Received %s %d/%s'
            % (self.ws_id, frame.uid, opcode, message))
        await self.message_handler(opcode, message, wss, self)

  async def broadcast(self, msg, opcode=OPCODES.TEXT, only=None):
    msg = bytes(msg, 'utf-8') if (isinstance(msg, str)) else msg
    dead = []
    for k in self.streams:
      wss = self.streams[k]
      if (only is None) or only(wss):
        try:
          await wss.send(msg, opcode)
        except (KeyError, OSError):
          dead.append(k)
    for k in dead:
      del self.streams[k]


class WebsocketStream(object):
  HEADER_FIN = (1 << 7)
  HEADER_OPC = 0xf

  MASKING_BIT = (1 << 7)
  LENGTH_MASK = (0xef)

  LENGTH_7 = 0x7e
  LENGTH_16 = 1 << 16

  ZERO_MASK = b'\0\0\0\0'

  def __init__(self, conn, frame, env, uPK, mask=ZERO_MASK):
    self.uPK = uPK
    self.conn = conn
    self.frame = frame
    self.env = env
    self.mask = mask
    self.buffer = b''

  # Convenience properties
  uid = property(lambda s: s.frame.uid)
  remote_ip = property(lambda s: s.frame.remote_ip)

  def _apply_mask(self, mask, data):
    if mask == self.ZERO_MASK:
      return data
    for i in range(len(data)):
      data[i] ^= mask[i % 4]
    return data

  async def send(self, msg, opcode=OPCODES.TEXT):
    msg = bytes(msg, 'utf-8') if (isinstance(msg, str)) else msg

    masking = self.MASKING_BIT if (self.mask != self.ZERO_MASK) else 0

    buf = bytearray(b'01')
    buf[0] = (self.HEADER_FIN | opcode)
    buf[1] = masking

    length = len(msg)
    if length < self.LENGTH_7:
      buf[1] |= length
    elif length < self.LENGTH_16:
      buf[1] |= 0x7e
      buf += struct.pack('!H', length)
    else:
      buf[1] |= 0x7f
      buf += struct.pack('!Q', length)

    if masking:
      buf += self.mask
      buf += self._apply_mask(self.mask, bytearray(msg))
    else:
      buf += msg

    buf = bytes(buf)

    if self.uPK.trace:
      self.uPK.trace('[ws] Send %s %d/%s' % (self.frame.uid, opcode, buf))

    await self.conn.reply(self.frame, buf, eof=False)

  def process(self, data):
    self.buffer += data
    try:
      opcode, message, offset = None, b'', 0
      while True:
        fin, opc, data, offset = self.extract_frame(offset)
        if opc in (OPCODES.TEXT, OPCODES.BINARY, OPCODES.CONT):
          if opc != OPCODES.CONT:
            opcode = opc
          message += data
          if fin:
            yield (opcode, message)
            self.buffer = self.buffer[offset:]
            opcode, message, offset = None, b'', 0
      else:
        print('FIXME: handle control frame %s' % opc)
    except (KeyError, IndexError):
      pass

  def extract_frame(self, base):
    b0 = self.buffer[base]
    b1 = self.buffer[base+1]
    masking = (b1 & self.MASKING_BIT)
    length = (b1 & self.LENGTH_MASK)

    offset = 2
    if length == 0x7e:
      length = struct.unpack('!H', self.buffer[base+2:base+4])
      offset = 4
    elif length == 0x7f:
      length = struct.unpack('!Q', self.buffer[base+2:base+10])
      offset = 10

    if masking:
      mask = self.buffer[base+offset:base+offset+4]
      offset += 4
    else:
      mask = self.ZERO_MASK

    end = base+offset+length
    result = (
      (b0 & self.HEADER_FIN),
      (b0 & self.HEADER_OPC),
      self._apply_mask(mask, bytearray(self.buffer[base+offset:end])),
      end)

    return result



# Copyright (C) 2020, The Beanstalks Project ehf. and Bjarni R. Einarsson.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Commercial licenses are for sale. See the files README.md and COPYING.txt
# for more details.
#
### # #
#
# A low-level PageKite protocol implementation.
#
# Note that some of this code is not currently thread-safe, in particular
# the uPageKiteDefaults.connect() method should not run be run concurrently
# with the same set of Kite objects.
#
import gc
import random
import re
import sys
import struct
import time
import select
from hashlib import sha1
from struct import unpack


# This is a cache of DNS hints we have recived from the network.
_DNS_HINTS = {}

# A global counter for bytes we've sent over the network
SENT_COUNTER = 0
SENT_DELAYS = 0


try:
    import uasyncio as asyncio
    IS_MICROPYTHON = True
except ImportError:
    import asyncio
    IS_MICROPYTHON = False

try:
    SELECT_POLL_IN = (select.POLLPRI | select.POLLIN)
except AttributeError:
    SELECT_POLL_IN = (select.POLLIN)

try:
    from sys import print_exception as print_exc
except ImportError:
    from traceback import print_exc as _print_exc
    def print_exc(e):
      return _print_exc()

try:
  from time import ticks_ms
except ImportError:
  def ticks_ms():
    return int(time.time() * 1000)

try:
  from os import urandom as random_bytes
except ImportError:
  def random_bytes(length):
    return bytes(random.getrandbits(8) for r in range(0, length))

try:
  from os import ilistdir
except ImportError:
  from os import listdir
  def ilistdir(path):
    return ((d, None, None) for d in listdir(path))

try:
  import ubinascii
  def sha1hex(data):
    return str(ubinascii.hexlify(
      sha1(bytes(data, 'latin-1')).digest()), 'latin-1')
except ImportError:
  def sha1hex(data):
    return sha1(bytes(data, 'latin-1')).hexdigest()

try:
    import ussl as ssl
except ImportError:
    try:
        import ssl
    except ImportError:
        ssl = False

try:
    from utime import sleep_ms as real_sleep_ms
    from uasyncio import sleep_ms
except ImportError:
    def real_sleep_ms(ms):
        time.sleep(ms/1000.0)
    async def sleep_ms(ms):
        await asyncio.sleep(ms/1000.0)

try:
  import usocket as socket
  IOError = OSError
  async def sock_connect_stream(uPK, addr, ssl_wrap=False, timeouts=(20, 300)):
    if uPK.trace:
      uPK.trace('>>connect(%s, ssl_wrap=%s)' % (addr, ssl_wrap))
    s = socket.socket()
    try:
      s.settimeout(timeouts[0])
      await fuzzy_sleep_ms()
      s.connect(addr)
      s.settimeout(timeouts[1])
      await fuzzy_sleep_ms()
      if ssl_wrap:
        uPK.GC_COLLECT()
        await fuzzy_sleep_ms(30)
        return (s, ssl.wrap_socket(s))
      await fuzzy_sleep_ms()
      return (s, s)
    except:
      s.close()
      raise
except ImportError:
  import socket
  async def sock_connect_stream(uPK, addr, ssl_wrap=False, timeouts=(20, 300)):
    if uPK.trace:
      uPK.trace('>>connect(%s, ssl_wrap=%s)' % (addr, ssl_wrap))
    s = socket.socket()
    s.settimeout(timeouts[0])
    await fuzzy_sleep_ms()
    s.connect(addr)
    s.settimeout(timeouts[1])
    if ssl_wrap:
      await fuzzy_sleep_ms(30)
      s = ssl.wrap_socket(s)
    await fuzzy_sleep_ms()
    return (s, s.makefile("rwb", 0))


def upk_open(path, mode='r'):
  for root in ('/bootstrap_live', '/bootstrap', ''):
    try:
      return open(root + path, mode)
    except OSError:
      pass
  raise


class RejectedError(ValueError):
  pass

class EofTunnelError(IOError):
  pass


async def fuzzy_sleep_ms(ms=0):
  real_sleep_ms(1)
  if ms > 1:
    await sleep_ms(ms-1)
  return max(ms, 1)


class Kite:
  def __init__(self, name, secret, proto='http', handler=None):
    self.proto = proto
    self.name = name
    self.secret = secret
    self.challenge = ''
    self.handler = handler

  def __str__(self):
    return '%s://%s' % (self.proto, self.name)


class Frame:
  def __init__(self, uPK, data=None, headers=None, payload=None, cid=''):
    self.uPK = uPK
    self.cid = '%s' % (cid,)
    if data:
      hdr_len = data.index(b'\r\n\r\n')
      hdr = str(data[:hdr_len], 'latin-1')
      self.payload = data[hdr_len+4:]
      self.headers = dict(ln.strip().split(': ', 1)
        for ln in hdr.splitlines())
    else:
      self.headers = headers
      self.payload = payload

  eof = property(lambda s: s.headers.get('EOF', ''))
  sid = property(lambda s: s.headers.get('SID'))
  uid = property(lambda s: s.cid + s.headers.get('SID'))
  host = property(lambda s: s.headers.get('Host'))
  port = property(lambda s: s.headers.get('Port'))
  proto = property(lambda s: s.headers.get('Proto'))
  remote_ip = property(lambda s: s.headers.get('RIP'))
  ping = property(lambda s: s.headers.get('PING'))


class uPageKiteDefaults:
  APPNAME = 'uPageKite'
  APPURL = 'https://github.com/pagekite/upagekite'
  APPVER = '0.2.0u'
  PARSE_HTTP_HEADERS = re.compile('^(Co(n(nection|tent-\S+)|okie)|Host|Origin|Sec-WebSocket-\S+|U(pgrade|ser-Agent)):')
  FE_NAME = 'fe4_100.b5p.us'  # pagekite.net IPv4 pool for pagekite.py 1.0.0
  FE_PORT = 443
  DDNS_URL = ('http', 'up.pagekite.net',  # FIXME: https if enough RAM?
              '/?hostname=%(domain)s&myip=%(ips)s&sign=%(sign)s')
  TOKEN_LENGTH = 36
  WITH_SSL = (ssl is not False)

  # These are used to compensate for Micropython/ESP32 having
  # an overly dumb DNS implementation.
  FE_HINT_NAME = 'fe.b5p.us'
  FE_HINT_URL = ('http', 'pagekite.net', '/logs/relays.txt')

  TICK_INTERVAL = 16
  MIN_CHECK_INTERVAL = 16
  MAX_CHECK_INTERVAL = 900
  WATCHDOG_TIMEOUT = 60000
  TUNNEL_TIMEOUT = 240
  MAX_POST_BYTES = 64 * 1024
  RANDOM_PING_VALUES = False

  # These values are critical magic under MicroPython - if the balance
  # is wrong, we get garbage reads from our tunnel socket or run out of
  # memory on the ESP32. In particular, the send window may not be much
  # larger than FILE_READ_BYTES; so those must be raised together and
  # that will in turn effect RAM usage. This limits performance.
  SEND_WINDOW_BYTES = (1500 if IS_MICROPYTHON else 16384)
  FILE_READ_BYTES = (1500 if IS_MICROPYTHON else 4*1500) - 64
  MS_DELAY_PER_BYTE = (0.025 if IS_MICROPYTHON else 0.005)

  WEBSOCKET_MASK = lambda: b'\0\0\0\0'
  WEBSOCKET_MAX_CONNS = (5 if IS_MICROPYTHON else 100)

  # Set to lambda: None to disable
  GC_COLLECT = (gc.collect if IS_MICROPYTHON else (lambda: None))
  trace = False  # Set to log in subclass to enable noise
  debug = False  # Set to log in subclass to enable noise
  info = False   # Set to log in subclass to enable noise
  error = False  # Set to log in subclass to enable noise

  @classmethod
  def log(cls, message):
    print('[%d] %s' % (ticks_ms(), message))

  @classmethod
  def addr_to_quad(cls, addr):
    if isinstance(addr, tuple):
      return addr[0]
    if isinstance(addr, bytearray):
      return '.'.join('%d' % b for b in unpack('8B', addr)[4:])
    return addr

  @classmethod
  def make_random_secret(cls, salt=''):
    # We do not know how good our randomness is; if it is really good, then
    # the salt and platform and time don't matter. If it is bad, then we
    # hope xor'ing against the SHA1 of secret values will compensate.
    salt = sha1(bytes(
        '%s' % ((ticks_ms(), sys.platform, sys.implementation, salt),),
        'latin-1'
      )).digest()
    rs = bytes((b^salt[i]) for i, b in enumerate(random_bytes(16)))
    if cls.trace:
      cls.trace('Random secret: %s' % rs)
    return rs

  @classmethod
  async def network_send_sleep(uPK, sent):
    global SENT_COUNTER
    global SENT_DELAYS

    SENT_COUNTER += sent
    if SENT_COUNTER > 100*1024*1024:
      SENT_COUNTER = SENT_DELAYS = 0

    sleep_time = 5 + int((SENT_COUNTER * uPK.MS_DELAY_PER_BYTE) - SENT_DELAYS)
    SENT_DELAYS += sleep_time
    await fuzzy_sleep_ms(sleep_time)

  @classmethod
  async def check_fe_hint_url(cls):
    global _DNS_HINTS
    result = None
    try:
      proto, host, path = cls.FE_HINT_URL
      result = await cls.http_get(proto, host, path, dns_hints=True)
      if ' 200 ' not in result[0]:
        raise Exception('Failed')
      cls.scan_for_dns_hints(str(result[-1] or '', 'latin-1').splitlines())
      if cls.debug:
        cls.debug('Updated DNS hints: %s' % (_DNS_HINTS))
    except Exception as e:
      if cls.error:
        cls.error(
          'Failed to update DNS hints: %s => (%s, %s)'
          % (cls.FE_HINT_URL, result, e))

  @classmethod
  def scan_for_dns_hints(cls, lines):
    global _DNS_HINTS
    for line in lines:
      if 'X-DNS: ' == line[:7]:
        xdns, host, ips = line.split(' ', 2)
        ips = [i.strip() for i in ips.split(',')]
        if host and ips:
          _DNS_HINTS[host] = ips

  @classmethod
  async def http_get(cls, proto, http_host, path,
                     addr=None, atonce=1024, maxread=8196, dns_hints=False):
    if addr is None:
      addr = http_host
    if hasattr(addr, 'split'):
      if ':' in addr:
        hostname, port = addr.split(':')
      else:
        port = 443 if (proto == 'https') else 80
        hostname = addr
      await fuzzy_sleep_ms()
      addr = socket.getaddrinfo(hostname, int(port))[0][-1]

    conn = None
    try:
      await fuzzy_sleep_ms()
      t0 = ticks_ms()
      cfd, conn = await sock_connect_stream(cls, addr, ssl_wrap=(proto == 'https'))
      await cls.send(conn,
        'GET %s HTTP/1.0\r\nHost: %s\r\n\r\n' % (path, http_host))

      t1 = ticks_ms()
      response = b''
      while len(response) < maxread:
        try:
          await fuzzy_sleep_ms()
          data = conn.read(atonce)
          response += data
        except:
          data = None
        if not data:
          break
      conn.close()
      t2 = ticks_ms()

      try:
        header, body = response.split(b'\r\n\r\n', 1)
      except ValueError:
        header = response
        body = ''
      header_lines = (str(header, 'latin-1') or '\n').splitlines()
      if dns_hints:
        cls.scan_for_dns_hints(header_lines)
      return (
        header_lines[0],
        dict(l.split(': ', 1) for l in header_lines[1:]),
        t1-t0, t2-t1, body)
    finally:
      try:
        conn.close()
      except:
        pass

  @classmethod
  async def get_kite_addrinfo(cls, kite):
    try:
      await fuzzy_sleep_ms(5)
      return socket.getaddrinfo(kite.name, cls.FE_PORT, socket.AF_INET, socket.SOCK_STREAM)
    except IOError:
      return []

  @classmethod
  async def get_relays_addrinfo(cls):
    global _DNS_HINTS
    addrs = []
    if cls.FE_NAME:
      await fuzzy_sleep_ms(5)
      try:
        addrs = socket.getaddrinfo(cls.FE_NAME, cls.FE_PORT, socket.AF_INET, socket.SOCK_STREAM)
      except IOError:
        pass
      for name in (cls.FE_NAME, cls.FE_HINT_NAME):
        for ip in _DNS_HINTS.get(name, []):
          await fuzzy_sleep_ms()
          try:
            ai = socket.getaddrinfo(ip, cls.FE_PORT, socket.AF_INET, socket.SOCK_STREAM)
            addrs.extend(ai)
          except IOError:
            pass
    return addrs

  @classmethod
  async def ping_relay(cls, relay_addr, bias=1.0):
    if cls.RANDOM_PING_VALUES:
      return random.randint(100, 300)
    try:
      l1, hdrs, t1, t2, body = await cls.http_get(
        'http', 'ping.pagekite', '/ping', relay_addr,
        atonce=250, dns_hints=True)

      elapsed = t1 + t2
      biased = int(float(elapsed) * bias)
      if 'X-PageKite-Overloaded' in hdrs:
        biased += 250 if (bias == 1.0) else 50

      if cls.debug:
        cls.debug('Ping %s ok: %dms (~%dms)' % (relay_addr, elapsed, biased))
      return biased
    except (IOError, OSError) as e:
      if cls.info:
        cls.info('Ping %s failed: %s' % (relay_addr, e))
      return 99999

  @classmethod
  def sync_send(cls, conn, data):
    data = bytes(data, 'latin-1') if isinstance(data, str) else data
    if cls.trace:
      cls.trace(']> %s' % data)
    conn.write(data)
    if hasattr(conn, 'flush'):
      conn.flush()

  @classmethod
  async def send(cls, conn, data):
    data = bytes(data, 'latin-1') if isinstance(data, str) else data
    conn.write(data)
    await cls.network_send_sleep(len(data))
    if hasattr(conn, 'flush'):
      conn.flush()
    if cls.trace:
      cls.trace('>>[%d] %s' % (len(data), data[:24]))

  @classmethod
  def fmt_chunk(cls, data):
    data = bytes(data, 'latin-1') if isinstance(data, str) else data
    return b'%x\r\n%s' % (len(data), data)

  @classmethod
  def fmt_data(cls, frame, data):
    return cls.fmt_chunk(b'SID: %s\r\n\r\n%s'% (
      bytes(frame.sid, 'latin-1'),
      bytes(data, 'latin-1') if isinstance(data, str) else data))

  @classmethod
  def fmt_eof(cls, frame):
    return cls.fmt_chunk(b'SID: %s\r\nEOF: 1WR\r\n\r\n' % (
      bytes(frame.sid, 'latin-1'),))

  @classmethod
  def fmt_pong(cls, pong):
    return cls.fmt_chunk(b'NOOP: 1\r\nPONG: %s\r\n\r\n!' % (
      bytes(pong, 'latin-1') if isinstance(pong, str) else pong,))

  @classmethod
  def fmt_ping(cls, conn):
    return cls.fmt_chunk(b'NOOP: 1\r\nPING: %.2f\r\n\r\n!' % (time.time(),))

  @classmethod
  async def read_http_header(cls, conn):
    header = bytes()
    read_bytes = 1
    await fuzzy_sleep_ms(20)
    while header[-4:] != b'\r\n\r\n':
      byte = conn.read(1)
      if byte in (b'', None):
        raise EofTunnelError()
      header += byte
    if cls.trace:
      cls.trace('<< %s' % header)
    await fuzzy_sleep_ms()
    return header

  @classmethod
  async def read_chunk(cls, conn):
    hdr = b''
    try:
      while not hdr.endswith(b'\r\n') and len(hdr) < 10:
        byte = conn.read(1)
        if byte in (b'', None):
          raise EofTunnelError()
        hdr += byte

      chunk_len = int(str(hdr, 'latin-1').strip(), 16)
      payload = b''
      while len(payload) < chunk_len:
        payload += conn.read(chunk_len - len(payload))

      if len(payload) != chunk_len:
        raise EofTunnelError(
          'Read size mismatch: %s != %s' % (len(payload), chunk_len))
      if cls.trace:
        cls.trace('<<[%d] %s %s' % (len(payload)+len(hdr), hdr, payload[:40]))

      return payload
    except (UnicodeError, ValueError):
      if cls.debug:
        cls.debug('Invalid chunk header: %s' % hdr)
      raise EofTunnelError()
    except OSError:
      raise EofTunnelError()

  @classmethod
  def sign(cls, secret, payload, salt=None, ts=None, length=None):
    if not salt:
      salt = '%x' % struct.unpack('I', cls.make_random_secret()[:4])
    if not length:
      length = cls.TOKEN_LENGTH
    if ts:
      salt = 't' + salt[1:]
      payload += '%x' % int(ts / 600)
    return (
      salt[0:8] +
      sha1hex(secret + payload + salt[0:8])[:length-8])

  @classmethod
  def x_pagekite(cls, relay_addr, kites, global_secret):
    reqs = []
    for kite in kites:
      client_token = sha1hex(
          '%s/%s/%s' % (global_secret, relay_addr, kite.secret)
        )[:cls.TOKEN_LENGTH]
      server_token = kite.challenge or ''
      data = '%s:%s:%s:%s' % (kite.proto, kite.name, client_token, server_token)
      sign = cls.sign(kite.secret, data)
      reqs.append('X-PageKite: %s:%s\r\n' % (data, sign))
    return ''.join(reqs)

  @classmethod
  def parse_challenge(cls, challenge, kites):
    ok = []
    rejected = []
    needsign = []
    for line in str(challenge, 'latin-1').splitlines():
      if line.startswith('X-PageKite-SignThis:'):
        parts = line.split(':')
        proto = parts[1].strip()
        kitename = parts[2]
        for kite in kites:
          if kite.name == kitename and kite.proto == proto:
            kite.challenge = parts[4]  # FIXME: Not thread safe!
            needsign.append(kite)
      elif line.startswith('X-PageKite-OK:'):
        ok.append(':'.join(line.split(':')[:3]))
      elif line.startswith('X-PageKite-Reject'):
        rejected.append(':'.join(line.split(':')[:3]))
      elif line.startswith('X-PageKite-Duplicate'):
        rejected.append(':'.join(line.split(':')[:3]))
      else:
        pass  # Parse the other lines too?
    return ok, needsign, rejected

  @classmethod
  async def connect(cls, relay_addr, kites, global_secret):
    if cls.debug:
      cls.debug('Flying %s via %s' % (
        ', '.join(str(k) for k in kites), relay_addr))

    for kite in kites:
      kite.challenge = ''

    # Connect, get fresh challenges
    cfd, conn = await sock_connect_stream(cls, relay_addr, ssl_wrap=cls.WITH_SSL)
    await cls.send(conn, (
        'CONNECT PageKite:1 HTTP/1.0\r\n'
        'X-PageKite-Features: AddKites\r\n'
        'X-PageKite-Version: %s\r\n'
        '%s\r\n'
      ) % (
        cls.APPVER,
        cls.x_pagekite(relay_addr, kites, global_secret)))

    # Make sense of it...
    challenge = await cls.read_http_header(conn)
    ok, needsign, rejected = cls.parse_challenge(challenge, kites)
    if rejected:
      conn.close()
      raise RejectedError(', '.join(rejected))

    if needsign:
      await cls.send(conn, cls.fmt_chunk((
          'NOOP: 1\r\n'
          '%s\r\n'
        ) % (
          cls.x_pagekite(relay_addr, needsign, global_secret))))
      challenge = await cls.read_chunk(conn)
      ok2, needsign, rejected = cls.parse_challenge(challenge, kites)
      ok += ok2
      if rejected or needsign:
        conn.close()
        raise RejectedError(', '.join(rejected + needsign))

    if not ok:
      conn.close()
      raise RejectedError('No requests accepted, is this really a relay?')

    if cls.info:
      cls.info('Connected to %s' % (relay_addr,))

    await fuzzy_sleep_ms()
    return cfd, conn

  @classmethod
  async def update_dns(cls, relay_ip, kites):
    proto, host, path_fmt = cls.DDNS_URL
    errors = 0
    for kite in kites:
      try:
        payload = '%s:%s' % (kite.name, relay_ip)
        l1, hdrs, t1, t2, body = await cls.http_get(
          proto, host, path_fmt % {
            'domain': kite.name,
            'sign': cls.sign(kite.secret, payload, length=100),
            'ips': relay_ip})
        if not (body.startswith(b'good') or body.startswith(b'nochg')):
          errors += 1
      except Exception as e:
        body = 'failed, %s' % e
        errors += 1
      if cls.debug:
        if not isinstance(body, str):
          body = str(body, 'latin-1')
        cls.debug('DNS update %s to %s: %s' % (
          kite.name, relay_ip, body.strip()))

    return (errors == 0)



# Copyright (C) 2020, The Beanstalks Project ehf. and Bjarni R. Einarsson.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Commercial licenses are for sale. See the files README.md and COPYING.txt
# for more details.
#
from .proto import socket


class DNSQuery:
  def __init__(self, data):
    self.query = data
    self.qdomain = ''
    if ((data[2] >> 3) & 15) == 0:
      ppos = 12
      plen = data[ppos]
      while plen:
        ppos += 1
        self.qdomain += str(data[ppos:ppos+plen], 'latin-1') + '.'
        ppos += plen
        plen = data[ppos]
    ppos += 1
    self.qend = ppos + 4
    self.qtype = ('IN A'
      if (data[ppos:self.qend] == b"\x00\x01\x00\x01")
      else 'unknown')

  def response(self, ip):
    if self.qtype == 'IN A':
      flags = b"\x81\x80"
      answers = b"\x00\x01"
      response = (b''
        + self.query[12:self.qend]  # Question, again
        + b"\x00\x00\x00\x01"       # TTL=1s
        + b"\x00\x04"               # RDLENGTH=4
        + ip)
    else:
      flags = b"\x81\x82"  # SERVFAIL
      answers = b"\x00\x00"
      response = b''

    return (b''
      + self.query[:2]            # Identification
      + flags
      + b"\x00\x01"               # Questions: 1
      + answers
      + b"\x00\x00\x00\x00"       # Auth: 0, Additional: 0
      + self.query[12:self.qend]  # Original question
      + response)


class CDNS:
  def __init__(self, ip, port, uPK):
    self.ip = bytes(map(int, ip.split(".")))
    self.uPK = uPK
    self.fd = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    self.fd.setblocking(False)
    self.fd.bind(socket.getaddrinfo('0.0.0.0', port)[0][-1])

  async def process_io(self):
    try:
      data, addr = self.fd.recvfrom(4096)
      query = DNSQuery(data)
      if query.qdomain:
        if self.uPK.info:
          self.uPK.info('[dns] Responding to %s query from %s for %s'
            % (query.qtype, addr[0], query.qdomain))
        self.fd.sendto(query.response(self.ip), addr)
      elif self.uPK.debug:
        self.uPK.debug('[dns] Unparsed query from %s: %s' % (addr[0], data))

    except Exception as e:
      print('Oops in CDNS: %s' % e)

    return True



# Copyright (C) 2020, The Beanstalks Project ehf. and Bjarni R. Einarsson.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Commercial licenses are for sale. See the files README.md and COPYING.txt
# for more details.
#
import gc
import time
import select

from .proto import asyncio, socket, ticks_ms, fuzzy_sleep_ms, print_exc
from .proto import Kite, Frame, EofTunnelError, uPageKiteDefaults
from .proto import SELECT_POLL_IN


try:
  from ntptime import settime
except ImportError:
  settime = None

def _ntp_settime(uPK):
  if settime is not None:
    try:
      if uPK.info:
        uPK.info('Attempting to set the time using NTP...')
      settime()
    except Exception as e:
      if uPK.error:
        uPK.error('Failed to set NTP time: %s' % e)


class LocalHTTPKite(Kite):
  def __init__(self, listen_on, name, secret, handler):
    Kite.__init__(self, name, secret, 'http', handler)
    self.listening_port = listen_on
    self.handlers = {}
    try:
      self.fd = socket.socket()
      self.fd.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
      self.fd.bind(socket.getaddrinfo('0.0.0.0', listen_on)[0][-1])
      self.fd.listen(5)
    except Exception as e:
      print_exc(e)
      print("Oops, binding socket failed: %s" % e)
      self.fd = None
    self.conns = {}

  def __str__(self):
    return '<LocalHTTPKite(%s://%s):%d>' % (
      self.proto, self.name, self.listening_port)

  def sync_reply(self, frame, data=None, eof=True):
    try:
      sock, client = self.conns[frame.sid]
    except KeyError:
      if data or not eof:
        print('**BUG?**  sync_reply(sid=%s), no conn found.' % frame.sid)
        raise
    if data:
      data = bytes(data, 'latin-1') if (isinstance(data, str)) else data
      sock.setblocking(True)
      client.write(data)
      if hasattr(client, 'flush'):
        client.flush()
    if eof:
      if frame.sid in self.handlers:
        del self.handlers[frame.sid]
      if frame.sid in self.conns:
        del self.conns[frame.sid]
      try:
        client.close()
      except:
        pass

  async def reply(self, frame, data=None, eof=True):
    await fuzzy_sleep_ms()
    self.sync_reply(frame, data=data, eof=eof)
    await frame.uPK.network_send_sleep(len(data))

  def await_data(self, uPK, sid, handler, nbytes=-1):
    async def async_handler(*args):
      return handler(*args)
    self.async_await_data(uPK, sid, async_handler, nbytes=-1)

  def async_await_data(self, uPK, sid, handler, nbytes=-1):
    # FIXME: We should be using asyncio sockets here, or add to our own
    #        event loop instead of this ugly hack. But it works, so...
    async def async_wait_job(nbytes):
      sock = client = None
      try:
        sock, client = self.conns[sid]
        poller = select.poll()
        poller.register(sock)
        while nbytes and (sid in self.handlers):
          readable = error = False
          for (obj, ev) in poller.poll(1):
            if (ev & SELECT_POLL_IN):
              readable = True
            if ev - (ev & (SELECT_POLL_IN|select.POLLOUT)):
              error = True
          uPK.GC_COLLECT()
          if readable:
            sock.setblocking(False)
            more = client.read(min(2048, nbytes if (nbytes > 0) else 2048))
            await handler(Frame(uPK, payload=more, headers={
              'SID': sid,
              'EOF': '1WR' if (not more) else ''}))
            if more:
              if nbytes > 0:
                nbytes -= len(more)
              await uPK.network_send_sleep(len(more))
            else:
              break
          elif error:
            break
          else:
            await fuzzy_sleep_ms(50)
      except Exception as e:
        if uPK.debug and (sid in self.conns):
          print_exc(e)
      finally:
        if sid in self.handlers:
          del self.handlers[sid]
        if sid in self.conns:
          del self.conns[sid]
        if client:
          client.close()
    self.handlers[sid] = True
    asyncio.get_event_loop().create_task(async_wait_job(nbytes))

  def close(self, sid=None):
    for _sid in ([sid] if (sid is not None) else list(self.conns.keys())):
      sock, client = self.conns.get(_sid, (None, None))
      if _sid in self.handlers:
        del self.handlers[_sid]
      if _sid in self.conns:
        del self.conns[_sid]
      if client is not None:
        client.close()
      elif sock is not None:
        sock.close()
    if sid is None:
      self.handlers = {}
      self.conns = {}
      self.fd.close()

  async def process_io(self, uPK):
    sid = sock = client = None
    try:
      sock, addr = self.fd.accept()
    except OSError as e:
      print_exc(e)
      return True

    try:
      sid = '%s-%x' % (sock.fileno(), ticks_ms())
      if hasattr(sock, 'makefile'):
        client = sock.makefile('rwb')
      else:
        client = sock
      self.conns[sid] = (sock, client)

      if uPK.trace:
        uPK.trace('Reading request from: %s' % (addr,))
      sock.setblocking(False)
      req = b''
      deadline = ticks_ms() + 500  # When to give up on data
      sleep_t = 20
      while (ticks_ms() < deadline) and (b'\r\n\r\n' not in req):
        await fuzzy_sleep_ms(min(sleep_t, 50))
        sleep_t += 5
        try:
          data = client.read(2048)
          if data is not None:
            req += data
        except socket.error:
          pass

      if uPK.trace:
        uPK.trace('Got local request: %s' % (req,))
      if b'\r\n\r\n' in req:
        await self.handler(self, self, Frame(uPK, payload=req, headers={
          'SID': sid,
          'Host': '0.0.0.0',
          'Proto': 'http',
          'Port': self.listening_port,
          'RIP': '::ffff:%s' % (addr[0],)}))
        sid = None  # Avoid the close below
      else:
        client.write(b'HTTP/1.0 408 Timed out\r\n\r\n')

    except KeyboardInterrupt:
      raise
    except Exception as e:
      print_exc(e)
      print('Oops, process_io: %s' % e)
      return False
    finally:
      if sid:
        self.close(sid)

    return True


class myLock:
  def __init__(self):
    self.locked = False

  async def __aenter__(self, *args):
    while self.locked:
      await fuzzy_sleep_ms(2)
    self.locked = True

  async def __aexit__(self, *args):
    self.locked = False
    await fuzzy_sleep_ms(2)


class uPageKiteConn:
  def __init__(self, pk):
    self.pk = pk
    try:
      self.lock = asyncio.Lock()
    except:
      self.lock = myLock()

  async def connect(self, relay_addr):
    pk = self.pk
    self.ip = pk.uPK.addr_to_quad(relay_addr)
    self.fd, self.conn = await pk.uPK.connect(relay_addr, pk.kites, pk.secret)
    now = int(time.time())
    self.last_data_ts = now
    self.last_handle_ts = now
    self.handlers = {}
    return self

  def __str__(self):
    return "<uPageKiteConn(%s)>" % self.ip

  def sync_reply(self, frame, data=None, eof=True):
    uPK = self.pk.uPK
    if data:
      uPK.sync_send(self.conn, uPK.fmt_data(frame, data))
    if eof:
      uPK.sync_send(self.conn, uPK.fmt_eof(frame))

  async def reply(self, frame, data=None, eof=True):
    uPK = self.pk.uPK
    async with self.lock:
      if data:
        await uPK.send(self.conn, uPK.fmt_data(frame, data))
      if eof:
        await uPK.send(self.conn, uPK.fmt_eof(frame))

  async def send_ping(self):
    async with self.lock:
      await self.pk.uPK.send(self.conn, self.pk.uPK.fmt_ping())

  def await_data(self, uPK, sid, handler, nbytes=-1):
    async def async_handler(*args):
      return handler(*args)
    self.async_await_data(uPK, sid, async_handler, nbytes)

  def async_await_data(self, uPK, sid, handler, nbytes=None):
    self.handlers[sid] = handler

  def close(self, sid=None):
    if not sid:
      self.handlers = {}
      if self.conn is not None:
        self.conn.close()
        self.conn = None
      if self.fd is not None:
        self.fd.close()
        self.fd = None
    else:
      # FIXME: Send EOF over tunnel?
      if sid in self.handlers:
        del self.handlers[sid]

  async def process_io(self, uPK):
    try:
      async with self.lock:
        frame = await self.pk.uPK.read_chunk(self.conn)
      now = int(time.time())
      self.last_data_ts = now
      if frame:
        frame = Frame(uPK, frame, cid=('%d-' % self.fd.fileno()))

        if frame.ping:
          async with self.lock:
            # Should never happen, as we send our own pings much more frequently
            await self.pk.uPK.send(self.conn, self.pk.uPK.fmt_pong(frame.ping))

        elif frame.sid and frame.sid in self.handlers:
          try:
            self.last_handle_ts = now
            await self.handlers[frame.sid](frame)
          except Exception as e:
            print_exc(e)
            print('Oops, sid handler: %s' % e)
            await self.reply(frame, eof=True)
            if frame.sid in self.handlers:
              del self.handlers[frame.sid]

        elif frame.sid and frame.host and frame.proto:
          for kite in self.pk.kites:
            if kite.name == frame.host and kite.proto == frame.proto:
              # FIXME: We should allow the handler to return a callback
              #        for any subsequent data with the same SID, to
              #        allow for uploads or bidirectional comms.
              self.last_handle_ts = now
              await kite.handler(kite, self, frame)
              break

        elif frame.sid:
          await self.reply(frame, eof=True)

        # FIXME: Detect and report quota values? Other things?

      # Zero-length chunks aren't an error condition
      return True
    except EofTunnelError as e:
      if self.pk.uPK.debug:
        self.pk.uPK.debug('EOF tunnel %s(%s)' % (type(e), e))
      return False


class uPageKiteConnPool:
  def __init__(self, conns, pk):
    self.pk = pk

    self.conns = dict((c.fd.fileno(), c) for c in conns)
    for so in pk.socks:
      if so.fd is not None:
        self.conns[so.fd.fileno()] = so

    self.poll = select.poll()
    for fno in self.conns:
      self.poll.register(self.conns[fno].fd, select.POLLIN)

  async def async_poll(self, timeout_ms):
    deadline = ticks_ms() + timeout_ms
    while ticks_ms() < deadline:
      events = self.poll.poll(1)
      if events:
        if self.pk.uPK.trace:
          self.pk.uPK.trace('poll() returned: %s' % (events,))
        return events
      await fuzzy_sleep_ms(min(75, deadline - ticks_ms()))
    return []

  async def process_io(self, uPK, timeout_ms):
    count = 0
    if self.pk.uPK.trace:
      self.pk.uPK.trace('Entering poll(%d)' % timeout_ms)

    for (obj, event) in await self.async_poll(timeout_ms):
      conn = self.conns[obj if (type(obj) == int) else obj.fileno()]
      if self.pk.uPK.trace:
        self.pk.uPK.trace(
          'process_io(%s) o=%s ev=0x%x' % (conn, obj, event))
      if (event & SELECT_POLL_IN):
        if await conn.process_io(uPK):
          count += 1
        else:
          self.pk.uPK.debug('conn.process_io() returned False')
          return False
      elif event & select.POLLOUT:
        pass
      else:
        return False

    if count == 0:
      dead = int(time.time()) - max(
        self.pk.uPK.MIN_CHECK_INTERVAL * 6,
        self.pk.uPK.TUNNEL_TIMEOUT)
      for conn in self.conns.values():
        if hasattr(conn, 'send_ping'):
          if conn.last_data_ts < dead:
            if self.pk.uPK.info:
              self.pk.uPK.info(
                'No PING response from %s, assuming it is down.' % (conn,))
            return False
          elif conn.last_data_ts < dead + (self.pk.uPK.MIN_CHECK_INTERVAL * 2):
            await conn.send_ping()

    return count


class uPageKite:
  def __init__(self, kites, socks=[], uPK=uPageKiteDefaults, public=True):
    self.uPK = uPK
    self.keep_running = True
    self.public = public
    self.kites = kites
    self.socks = [s for s in socks if s.fd is not None]
    self.secret = uPK.make_random_secret([(k.name, k.secret) for k in kites])
    self.want_dns_update = [0]
    self.reconfig_flag = False

  def reconfigure(self):
    self.reconfig_flag = True
    return self

  async def choose_relays(self, preferred=[]):
    await fuzzy_sleep_ms()

    relays = []
    if len(self.kites) == 0:
      return relays

    await self.uPK.check_fe_hint_url()

    for kite in self.kites:
      for a in await self.uPK.get_kite_addrinfo(kite):
        if a[-1] not in relays and len(relays) < 10:
          relays.append(a[-1])

    if self.kites:
      for a in await self.uPK.get_relays_addrinfo():
        if a[-1] not in relays and len(relays) < 10:
          relays.append(a[-1])
    if not relays:
      if self.kites and self.uPK.info:
        self.uPK.info('No relays found in DNS, is our Internet down?')
      return []

    if len(relays) == 1:
      return relays

    pings = [0] * len(relays)
    for i, relay_addr in enumerate(relays):
      bias = 0.9 if (not i or relay_addr in preferred) else 1.0
      pings[i] = await self.uPK.ping_relay(relay_addr, bias)

    relays = list(zip(pings, relays))
    fastest = relays[0]
    for r in relays:
      if r[0] < fastest[0]:
        fastest = r

    if fastest != relays[0]:
      return [fastest[-1], relays[0][-1]]
    else:
      return [relays[0][-1]]

  async def connect_relays(self, relays, now):
    conns = []
    self.want_dns_update = [0]
    for relay in relays:
      await fuzzy_sleep_ms()
      try:
        conns.append(await uPageKiteConn(self).connect(relay))
      except KeyboardInterrupt:
        raise
      except Exception as e:
        if self.uPK.error:
          self.uPK.error('Failed to connect %s: %s' % (relay, e))
    if conns:
      self.want_dns_update = [now - 1, conns[0].ip]
    return conns

  async def relay_loop(self, conns, deadline):
    max_timeout = 5005
    wdt = None
    try:
      if self.uPK.WATCHDOG_TIMEOUT:
        from machine import WDT
        wdt = WDT(timeout=self.uPK.WATCHDOG_TIMEOUT)
        max_timeout = min(max_timeout, self.uPK.WATCHDOG_TIMEOUT // 2)
    except KeyboardInterrupt:
      raise
    except:
      pass

    await fuzzy_sleep_ms()
    try:
      pool = uPageKiteConnPool(conns, self)
      while pool.conns and time.time() < deadline:
        if wdt:
          wdt.feed()

        self.uPK.GC_COLLECT()
        timeout_ms = min(max(100, (deadline - time.time()) * 1000), max_timeout)
        await fuzzy_sleep_ms()
        if await pool.process_io(self.uPK, int(timeout_ms)) is False:
          raise EofTunnelError('process_io returned False')

        if self.reconfig_flag:
          if self.uPK.info:
            self.uPK.info("Exiting relay_loop early: reconfiguration requested.")
          return False

      # Happy ending!
      return True
    except KeyboardInterrupt:
      raise
    except Exception as e:
      print_exc(e)
      print('Oops, relay_loop: %s(%s)' % (type(e), e))

    # We've fallen through to our unhappy ending, clean up
    for conn in conns:
      try:
        conn.close()
      except Exception as e:
        print("Oops, close(%s): %s" % (conn, e))
    return False

  # This is easily overridden by subclasses
  async def tick(self, **attrs):
    if self.uPK.info:
      self.uPK.info("Tick: %s %s%s"
        % (self.uPK.APPNAME, self.uPK.APPVER,
           ''.join('; %s=%s' % pair for pair in attrs.items())))

  async def check_relays(self, now, back_off):
    self.want_dns_update = [0]
    relays = await self.choose_relays()
    if relays:
      relays = await self.connect_relays(relays, now)
      # FIXME: Did we fail to make some connections?

    if relays:
      back_off = 1
    else:
      back_off = min(back_off * 2,
        self.uPK.MAX_CHECK_INTERVAL // self.uPK.MIN_CHECK_INTERVAL)
      if self.uPK.info:
        self.uPK.info(
          "Next connection attempt in %d+ seconds..."
          % (back_off * self.uPK.MIN_CHECK_INTERVAL))

    return relays, back_off

  async def check_dns(self, now, relays, back_off):
    recheck_max = 0
    if now > 3600:
      recheck_max = max(0, 3600 // self.uPK.MIN_CHECK_INTERVAL)
      if 1 < self.want_dns_update[0] <= recheck_max:
        self.want_dns_update[0] -= 1
    else:
      pass  # Clock is wonky, disable the recheck magic for now

    if 1 == self.want_dns_update[0]:
      self.want_dns_update[0] = recheck_max
      for kite in self.kites:
        if self.uPK.trace:
          self.uPK.trace("Checking current DNS state for %s" % kite)
        for a in await self.uPK.get_kite_addrinfo(kite):
          if a[-1][0] not in self.want_dns_update:
            if self.uPK.info:
              self.uPK.info(
                "DNS for %s is wrong (%s), should update" % (kite, a[-1][0]))
            self.want_dns_update[0] = now + (back_off * self.uPK.MIN_CHECK_INTERVAL * 2)

      if (self.want_dns_update[0] == recheck_max) and (len(relays) > 1):
        if self.uPK.debug:
          self.uPK.debug("DNS is good, checking for unused relays.")
        idle_ts = now - (self.uPK.MAX_CHECK_INTERVAL * 2)
        idle = [r for r in relays[1:] if r.last_handle_ts < idle_ts]
        if idle:
          relays = [r for r in relays if r not in idle]
          for r in idle:
            if self.uPK.info:
              self.uPK.info("Disconnecting from relay: %s" % r)
            r.close()

    if recheck_max < self.want_dns_update[0] < now:
      if await self.uPK.update_dns(self.want_dns_update[1], self.kites):
        self.want_dns_update[0] = recheck_max
        if relays:
          back_off = 1
      else:
        back_off = min(back_off * 2,
          self.uPK.MAX_CHECK_INTERVAL // self.uPK.MAX_CHECK_INTERVAL)
        self.want_dns_update[0] = now + (back_off * self.uPK.MIN_CHECK_INTERVAL * 2)
        if self.uPK.info:
          self.uPK.info(
            "Next DNS update attempt in %d+ seconds..."
            % (self.want_dns_update[0] - now))

    return relays, back_off

  async def main(self):
    if time.time() < 0x27640000:
      _ntp_settime(self.uPK)

    next_check = int(time.time())
    back_off = 1
    relays = []
    while self.keep_running:
      self.reconfig_flag = False
      now = int(time.time())
      await self.tick(
        back_off=back_off,
        relays=len(relays),
        socks=len(self.socks),
        mem_free=(gc.mem_free() if hasattr(gc, 'mem_free') else 'unknown'))

      await fuzzy_sleep_ms()
      if next_check <= now and self.public:
        self.uPK.GC_COLLECT()

        # Check our relay status? Reconnect?
        if not relays:
          relays, back_off = await self.check_relays(now, back_off)
          await fuzzy_sleep_ms()

        # Is DNS up to date?
        if relays:
          relays, back_off = await self.check_dns(now, relays, back_off)
          await fuzzy_sleep_ms()

      # Schedule our next check; it should be neither too far in the future,
      # nor in the distant past. Clocks can misbehave, so we check for both.
      if next_check > now + back_off * self.uPK.MIN_CHECK_INTERVAL:
        next_check = now + back_off * self.uPK.MIN_CHECK_INTERVAL
      while next_check <= now:
        next_check += back_off * self.uPK.MIN_CHECK_INTERVAL

      # Process IO events for a while, or sleep.
      if relays or self.socks:
        if not await self.relay_loop(relays, now + self.uPK.TICK_INTERVAL):
          if relays:
            # We had a working connection, it broke! Reconnect ASAP.
            next_check = now
            for conn in relays:
              conn.close()
            relays = []
      else:
        if self.uPK.debug:
          self.uPK.debug("No sockets available, sleeping until %d" % next_check)
        await fuzzy_sleep_ms(999 * max(0, next_check - int(time.time())))

  def run(self):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(self.main())
    finally:
        loop.close()



# Copyright (C) 2020, The Beanstalks Project ehf. and Bjarni R. Einarsson.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Commercial licenses are for sale. See the files README.md and COPYING.txt
# for more details.

from .proto import asyncio


# Decorator for making functions handle POSTed data
def process_post(max_bytes=None, _async=False):
  def decorate(func):
    def post_wrapper(req_env):
      func_attrs = req_env['url_func_attrs']
      httpd = req_env['httpd']
      async def handler():
        try:
          await httpd.run_handler(func, func_attrs, req_env)
        except Exception as e:
          if httpd.uPK.debug:
            httpd.uPK.debug('post_wrapper failed: %s(%s)' % (type(e), e))
          req_env['send_http_response'](code=500, msg='Server Error')
      handle_big_request(handler, req_env, max_bytes=max_bytes, _async=True)
    if _async:
      async def async_post_wrapper(req_env):
        return post_wrapper(req_env)
      return async_post_wrapper
    else:
      return post_wrapper
  return decorate


def html_encode(txt):
  txt = txt.replace('&', '&amp;')
  txt = txt.replace('<', '&lt;')
  return txt.replace('>', '&gt;')


def form_encode(txt):
  txt = txt.replace('&', '&amp;')
  return '"%s"' % txt.replace('"', '&quot;')


def parse_hdr(hdr):
  # This parser is incorrect for foo="bar; baz"; type variables.
  if ';' in hdr:
    attrs = hdr.split(';')
    hdr = attrs.pop(0)
    attrs = dict(a.strip().split('=', 1) for a in attrs)
    for a in attrs:
      if attrs[a][:1] == attrs[a][-1:] == '"':
        attrs[a] = attrs[a][1:-1]
  else:
    attrs = {}
  return (hdr, attrs)


class PostVars(dict):
  def val(self, name, default=''):
    return self.get(name, {'value': default})['value']


class ParseNull():
  def __init__(self, uPK, frame, headers, attrs):
    self.uPK = uPK
    self.headers = headers
    self.frame = frame
    self.attrs = attrs
    self.parse()

  def parse(self):
    pass


class ParseWFUE(ParseNull):
  def parse(self):
    try:
      from .httpd import HTTPD
      if self.uPK.trace:
        self.uPK.trace('<<%s' % self.frame.payload)
      self.headers['_post_data'] = HTTPD.qs_to_list(
        str(self.frame.payload, 'utf-8'))
      self.frame.payload = b''
    except Exception as e:
      if self.uPK.debug:
        self.uPK.debug('%s parse failed: %s(%s)' % (self, type(e), e))


class ParseJSON(ParseNull):
  def parse(self):
    try:
      import json
      if self.uPK.trace:
        self.uPK.trace('<<%s' % self.frame.payload)
      self.headers['_post_data'] = json.loads(self.frame.payload)
      self.frame.payload = b''
    except Exception as e:
      if self.uPK.debug:
        self.uPK.debug('%s parse failed: %s(%s)' % (self, type(e), e))


def handle_big_request(handler, env, max_bytes=None, _async=False):
  uPK = env['httpd'].uPK
  frame = env['frame']
  if frame.payload:
    hdr, frame.payload = frame.payload.split(b'\r\n\r\n', 1)
    if uPK.trace:
      uPK.trace('<<%s' % (hdr + b'\r\n\r\n',))
    del hdr

  # Help our devs out a bit.
  if uPK.debug and max_bytes and (max_bytes > uPK.MAX_POST_BYTES):
    uPK.debug('BUG: max_bytes(%d) > MAX_POST_BYTES(%d), ignoring max_bytes'
      % (max_bytes, uPK.MAX_POST_BYTES))

  # This has to happen before we let the parser run, as parsers often
  # modify the frame.payload: in-place mutations are how we cope with
  # our memory constraints.
  headers = env['http_headers']
  max_bytes = min(max_bytes or uPK.MAX_POST_BYTES, uPK.MAX_POST_BYTES)
  needed_bytes = int(headers.get('Content-Length', 0)) - len(frame.payload)
  if needed_bytes > max_bytes:
    env['send_http_response'](code=400, msg='Too big')
    return None

  (ctype, cattrs) = parse_hdr(headers.get('Content-Type', 'text/plain'))
  if ctype == 'application/json':
    parser_cls = ParseJSON
  elif ctype == 'application/x-www-form-urlencoded':
    parser_cls = ParseWFUE
  elif ctype == 'multipart/form-data':
    from web_mpfd import ParseMPFD
    parser_cls = ParseMPFD
  else:
    parser_cls = ParseNull

  headers['_post_data'] = PostVars()
  parser = parser_cls(uPK, frame, headers, cattrs)
  uPK.GC_COLLECT()
  if not needed_bytes:
    del parser
    if _async:
      asyncio.get_event_loop().create_task(handler())
    else:
      handler()
    return None

  conn = env['conn']
  needed_bytes = [needed_bytes]  # Scope hack, nonlocal does not work
  async def update_frame(frameN):
    nbytes = len(frameN.payload or '')
    frame.payload += frameN.payload
    needed_bytes[0] -= len(frameN.payload)
    del frameN

    parser.parse()
    uPK.GC_COLLECT()
    if nbytes < 1 or needed_bytes[0] < 1:
      if frame.sid in conn.handlers:
        del conn.handlers[frame.sid]
      if uPK.trace and frame.payload:
        uPK.trace('<<%s' % frame.payload)
      if _async:
        await handler()
      else:
        handler()

  conn.async_await_data(uPK, frame.sid, update_frame, needed_bytes[0])
  return None



# Copyright (C) 2020, The Beanstalks Project ehf. and Bjarni R. Einarsson.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Commercial licenses are for sale. See the files README.md and COPYING.txt
# for more details.
#
import os
import sys
import time
import json

from .proto import print_exc, asyncio, ilistdir, upk_open, fuzzy_sleep_ms

try:
  from uos import stat
  def size(fn):
    return stat(fn)[6]
except ImportError:
  from os.path import getsize as size


_HANDLERS = {}

_MIMETYPES = {
  'css': 'text/css',
  'gif': 'image/gif',
  'htm': 'text/html; charset=utf-8',
  'html': 'text/html; charset=utf-8',
  'jpg': 'image/jpeg',
  'jpeg': 'image/jpeg',
  'js': 'text/javascript',
  'json': 'application/json',
  'md': 'text/plain; charset=utf-8',
  'pdf': 'application/pdf',
  'txt': 'text/plain; charset=utf-8',
  'xml': 'application/xml',
  '_default': 'application/octet-stream'}


# Method for registering new file-extension -> MIME type mappings
def register_mime_extensions(**kwargs):
  for k in kwargs:
    _MIMETYPES[str(k).lower()] = kwargs[k]


# Decorator for registering functions as URL handlers
def url(*paths, **attrs):
  def decorate(func):
    for path in paths:
      _HANDLERS[path] = (func, attrs)
    return func
  return decorate


# Decorator for registering async functions as URL handlers
def async_url(*paths, **attrs):
  attrs['_async'] = True
  return url(*paths, **attrs)


# Helper to look up a mimetype
def filename_to_mimetype(fn):
  return _MIMETYPES.get(fn.rsplit('.', 1)[-1].lower(), _MIMETYPES['_default'])


# Helper for iterating over chunks of a buffer
def _buffer_byte_chunks(buf, chunksize):
  start, end = 0, min(len(buf), chunksize)
  while start < len(buf):
    data = buf[start:end]
    yield str(data, 'latin-1') if isinstance(data, str) else bytes(data)
    start, end = end, min(len(buf), end + chunksize)


# Helper to read file descriptor right up until the end
def _read_fd_iterator(fd, readsize, first_item=None):
  if first_item is not None:
    yield first_item
  while True:
    data = fd.read(readsize)
    if data:
      yield data
    else:
      break

# Helper class for navigating the request environment
class ReqEnv(dict):
  # Details about the client
  remote_ip = property(lambda s: s['frame'].remote_ip)

  # Details about the HTTP request
  post_vars = property(lambda s: dict(s['http_headers'].get('_post_data', [])))
  post_data = property(lambda s: s['http_headers'].get('_post_data', {}))
  query_vars = property(lambda s: dict(s['http_headers']['_qs']))
  query_tuples = property(lambda s: s['http_headers']['_qs'])
  request_path = property(lambda s: s['http_headers']['_path'])
  http_method = property(lambda s: s['http_headers']['_method'])
  http_headers = property(lambda s: s['http_headers'])
  http_host = property(lambda s: s['frame'].host)
  http_port = property(lambda s: s['frame'].port)
  payload = property(lambda s: s['frame'].payload)
  frame = property(lambda s: s['frame'])


class HTTPD:
  METHODS = (
    'GET',
    'HEAD',
    'POST')

  def __init__(self, name, webroot, env, uPK):
    self.uPK = uPK
    self.name = name
    self.webroot = webroot
    self.static_max_age = 3600
    self.base_env = env

  @classmethod
  def unquote(cls, quoted):
    if isinstance(quoted, str):
      quoted = bytes(quoted, 'latin-1')
    _in = quoted.split(b'%')
    _out = [_in[0]]
    for frag in _in[1:]:
      try:
        _out.extend([bytes([int(frag[:2], 16)], 'latin-1'), frag[2:]])
      except ValueError:
        _out.extend([b'%', frag])
    try:
      return str(b''.join(_out), 'utf-8')
    except UnicodeError:
      return quoted

  @classmethod
  def qs_to_list(cls, qs):
    if not qs:
      return []
    return [
      [cls.unquote(part) for part in pair.split('=', 1)]
      for pair in qs.split('&')]

  def http_response(self, code, msg, mimetype, ttl=None, hdrs={}):
    return (
        'HTTP/%s %d %s\r\n'
        'Server: %s\r\n'
        '%s%s%s\r\n'
      ) % (
        '1.1' if ('Upgrade' in hdrs) else '1.0',
        code, msg, self.uPK.APPURL,
        ('Content-Type: %s\r\n' % mimetype) if mimetype else '',
        ('Cache-Control: max-age=%d\r\n' % ttl) if ttl else '',
        ''.join('%s: %s\r\n' % (k, v) for k, v in hdrs.items()))

  def log_request(self, frame, method, path, code,
                  sent='-', headers={}, user='-'):
    if self.uPK.info:
      self.uPK.info('[www] %s %s - %s %s:%s%s - %s %s - %s' % (
        user, frame.remote_ip,
        method, frame.host, frame.port, path,
        code, sent,
        headers.get('User-Agent', headers.get('user-agent', '-'))))

  async def _err(self, code, msg, method, path, conn, frame, headers={}):
    self.log_request(frame, method, path, code, headers=headers)
    await conn.reply(frame, self.http_response(code, msg, 'text/plain')+msg+'\n')

  async def background_send(self,
        iterator, first_reply, conn, frame, method, path, hdrs, _close=[]):
    # Abort the upload if the remote end closes the connection
    saw_eof = [False]
    progress = [0]
    def beware_eof(frm):
      saw_eof[0] = saw_eof[0] or ('W' in frm.eof)
      if 'SKB' in frm.headers:
        progress[0] = int(frm.headers['SKB'])

    # Iteratively send our data
    async def async_send_data():
      sent = 0
      code = 200
      want_eof = True
      sent_first = False
      try:
        await fuzzy_sleep_ms(10)

        first_item = next(iterator)
        if method == 'HEAD':
          first_item['eof'] = True
          first_item['suppress_log'] = False
          first_reply(**first_item)
          want_eof = False
          sent_first = True
          return

        want_eof = first_item.get('eof', want_eof)
        first_item['eof'] = False
        sent = first_reply(**first_item)
        sent_first = True

        # Note: The order and type of sleeps here is important; if we do
        #       not give control back to the main event loop which reads
        # the PageKite tunnel, the "ack" packets may build up and cause
        # our ESP32 devices to run out of RAM and break the connection.
        conn.await_data(self.uPK, frame.sid, beware_eof)
        for app_data in iterator:
          for data in _buffer_byte_chunks(app_data, self.uPK.SEND_WINDOW_BYTES):
            await fuzzy_sleep_ms(5)
            if saw_eof[0]:
              break

            await conn.reply(frame, data, eof=False)
            sent += len(data)
            # Avoid buffer bloat
            while (progress[0]
                  and not saw_eof[0]
                  and (progress[0] * 1024) < (sent - self.uPK.SEND_WINDOW_BYTES)):
              await fuzzy_sleep_ms(50)
          if saw_eof[0]:
            break

      except Exception as e:
        code = '-'
        want_eof = sent_first
        if not sent_first:
          sent = first_reply(code=500, msg='Server Error', eof=True)
        if self.uPK.debug:
          print_exc(e)
          self.uPK.debug('Exception in async_send_data: %s(%s)' % (type(e), e))
      finally:
        if method != 'HEAD':
          self.log_request(frame, method, path, code, sent, hdrs)
        for fd in _close:
          fd.close()
        if frame.sid in conn.handlers:
          del conn.handlers[frame.sid]
        if want_eof:
          try:
            await conn.reply(frame, None, eof=True)
          except:
            pass

    asyncio.get_event_loop().create_task(async_send_data())
    await fuzzy_sleep_ms(1)

  async def run_handler(self, func, func_attrs, req_env):
    req_env['url_func_attrs'] = func_attrs
    if func_attrs.get('_async'):
      result = await func(req_env)
    else:
      result = func(req_env)

    if result is not None:
      if isinstance(result, dict):
        await fuzzy_sleep_ms(1)
        req_env['send_http_response'](**result)
      else:
        await self.background_send(result,
          req_env['send_http_response'],
          req_env['conn'],
          req_env['frame'],
          req_env['http_headers']['_method'],
          req_env['http_headers']['_path'],
          req_env['http_headers'])

  async def handle_http_request(self, kite, conn, frame):
    method = path = '-'
    self.uPK.GC_COLLECT()
    try:
      headers = frame.payload.split(b'\r\n\r\n', 1)[0]
      request, headers = str(headers, 'latin-1').split('\r\n', 1)
      method, path, http = request.split(' ', 2)
      if ('..' in path) or method not in self.METHODS:
        raise ValueError()
      del request
      if method != 'POST':
        frame.payload = b''
      qs = ''
      if '?' in path:
        path, qs = path.split('?', 1)
      headers = dict(
        l[:128].split(': ', 1) for l in headers.splitlines()
        if self.uPK.PARSE_HTTP_HEADERS.match(l))
    except Exception as e:
      return await self._err(400, 'Invalid request', method, path, conn, frame)

    func, func_attrs = _HANDLERS.get(path, (None, None))
    filename = self.webroot + path

    await fuzzy_sleep_ms()
    if func is None:
      try:
        ls = [l[0] for l in ilistdir(filename)]
        if 'index.py' in ls:
          filename = filename + '/index.py'
        elif 'index.html' in ls:
          filename = filename + '/index.html'
        del ls
      except:
        pass
      try:
        fd = open(filename, 'rb')
      except:
        try:
          filename = self.webroot + '/404.py'
          fd = open(filename, 'rb')
        except:
          return await self._err(404, 'Not Found', method, path, conn, frame, headers)
    else:
      fd = None

    postponed = []
    await fuzzy_sleep_ms()
    try:
      def first_reply(
            body='', mimetype='text/html; charset=utf-8',
            code=200, msg='OK', ttl=None, eof=True, hdrs={},
            suppress_log=False):
        rdata = self.http_response(code, msg, mimetype, ttl, hdrs)
        if body and method != 'HEAD':
          rdata += body
        conn.sync_reply(frame, rdata, eof=eof)
        if not suppress_log:
          sent = len(rdata) if eof else '-'
          self.log_request(frame, method, path, code, sent, headers)
        return len(rdata)

      def postpone_action(func, *args, **kwargs):
        postponed.append((func, args, kwargs))

      if func or filename.endswith('.py'):
        headers['_method'] = method
        headers['_path'] = path
        headers['_qs'] = self.qs_to_list(qs)
        req_env = {
          'time': time, 'os': os, 'sys': sys, 'json': json, 'open': upk_open,
          'httpd': self, 'kite': kite, 'conn': conn, 'frame': frame,
          'send_http_response': first_reply,
          'postpone_action': postpone_action,
          'http_headers': headers}
        req_env.update(self.base_env)
        if fd:
          await fuzzy_sleep_ms(25)
          code = str(fd.read(), 'utf-8')
          self.uPK.GC_COLLECT()
          exec(code, req_env)
        else:
          await fuzzy_sleep_ms()
          await self.run_handler(func, func_attrs, ReqEnv(req_env))
      else:
        filesize = size(filename)
        await self.background_send(
          _read_fd_iterator(fd, self.uPK.FILE_READ_BYTES,
            first_item={
              'hdrs': {'Content-Length': filesize},
              'suppress_log': (filesize < 102400),
              'mimetype': filename_to_mimetype(filename),
              'ttl': self.static_max_age}),
          first_reply, conn, frame, method, path, headers, _close=[fd])
        fd = None
    except Exception as e:
      if self.uPK.debug:
        print_exc(e)
        self.uPK.debug('Exception in handle_http_request: %s(%s)' % (type(e), e))
      return await self._err(500, 'Server Error', method, path, conn, frame, headers)
    finally:
      if fd is not None:
        fd.close()

    for f, a, kw in postponed:
      await fuzzy_sleep_ms()
      f(*a, **kw)



# Copyright (C) 2020, The Beanstalks Project ehf. and Bjarni R. Einarsson.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# Commercial licenses are for sale. See the files README.md and COPYING.txt
# for more details.
#
from .web import PostVars, ParseNull, parse_hdr

# This should be an efficient state machine for iterating over a buffer,
# yielding each individual line for processing. Line lengths are capped
# and the caller can define a list of strings to stop on.
def _lines(payload, stop=(), maxlen=250, eol=b'\n'):
  total = 0
  try:
    while True:
      end = min(maxlen, payload[0][total:].index(eol))  # Raises ValueError
      ln = payload[0][total:total+end+1]
      yield ln
      total += len(ln)
      if ln in stop:
        break
  except ValueError:
    pass
  payload[0] = payload[0][total:]


class ParseMPFD(ParseNull):
  TEMP_FN = 0

  def __init__(self, *args):
    self.in_header = True
    self.varname = None
    self.payload = {'value': ''}
    ParseNull.__init__(self, *args)  # Last, because parse() depends on the above

  def _tempfile(self):
    for path in ('/tmp', ''):
      try:
        temp_filename = '%s/upload_%d.tmp' % (path, ParseMPFD.TEMP_FN)
        fd = open(temp_filename, 'wb')
        ParseMPFD.TEMP_FN += 1
        ParseMPFD.TEMP_FN %= 4
        return {'temp_filename': temp_filename, 'fd': fd}
      except OSError:
        pass
    raise

  def parse(self):
    post_data = self.headers['_post_data']
    box = [self.frame.payload]
    for line in _lines(box):
      if self.uPK.trace:
        self.uPK.trace('<<%s' % line)

      if (line[:2] == b'--') and line[2:].startswith(bytes(self.attrs['boundary'], 'latin-1')):
        if 'fd' in self.payload:
          self.payload['bytes'] -= 2
          self.payload['fd'].close()
          del self.payload['fd']
        if self.varname:
          if self.payload['value'].endswith('\r\n'):
            self.payload['value'] = self.payload['value'][:-2]
          post_data[self.varname] = self.payload
        self.in_header = True
        self.payload = {'value': ''}

      elif self.in_header:
        if line in (b'\n', b'\r\n'):
          self.in_header = False

        elif line.startswith(b'Content-Disposition:'):
          hname, hval = str(line, 'utf-8').split(':', 1)
          hval, hattrs = parse_hdr(hval.strip())
          self.varname = hattrs['name']
          if 'filename' in hattrs:
            self.payload['value'] = hattrs['filename']
            self.payload['bytes'] = 0
            self.payload.update(self._tempfile())

      else:
        if 'bytes' in self.payload:
          self.payload['bytes'] += len(line)
        if 'fd' in self.payload:
          self.payload['fd'].write(line)
        else:
          self.payload['value'] += str(line, 'utf-8')

    self.frame.payload = box[0]



from distutils.core import setup
setup(
  name = 'upagekite',
  packages = ['upagekite'],
  version = '0.2.0',
  license='GPL-3.0',
  description = "Embeddable/IoT PageKite asyncio web server and framework",
  long_description = """\
NOTE: This ALPHA code - a work in progress! Testing uploads to PyPI.

A minimalistic HTTP server and web "micro-framework" for developing Internet
accessible (not just local/LAN) dynamic web apps. The app is exposed to the
web automatically using the PageKite relay protocol.

uPagekite supports HTTP, streaming file up/downloads and websockets. It is
compatible with asyncio and uasyncio event loops.

See the Tutorial for examples: https://github.com/pagekite/upagekite-tutorial

Developed and tested on CPython 3.x on Linux, and MicroPython on the ESP32.

Licensed under the GPLv3+ by default, proprietary licenses are available
from the author.""",
  author = 'Bjarni R. Einarsson',
  author_email = 'bre@pagekite.net',
  url = 'https://github.com/pagekite/upagekite',
  download_url = 'https://codeload.github.com/pagekite/upagekite/tar.gz/refs/tags/v0.2.0u',
  keywords = ['pagekite', 'http', 'websocket', 'esp32', 'micropython'],
  install_requires=[],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'License :: Other/Proprietary License',
    'Framework :: AsyncIO',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: Implementation :: CPython',
    'Programming Language :: Python :: Implementation :: MicroPython',
    'Topic :: Software Development :: Embedded Systems',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Libraries :: Application Frameworks',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Topic :: Internet :: WWW/HTTP :: HTTP Servers'])
